@@ -19,7 +19,8 @@ const QString API_SUBSCRIPTIONS = QStringLiteral("/api/v1/auth/subscriptions");
 const QString API_TOP = QStringLiteral("/api/v1/top");
 const QString API_TRENDING = QStringLiteral("/api/v1/trending");
 const QString API_VIDEOS = QStringLiteral("/api/v1/videos");
-const QString API_CHANNELS = QStringLiteral("/api/v1/channels/videos");
+const QString API_CHANNEL_VIDEOS = QStringLiteral("/api/v1/channels/videos");
+const QString API_CHANNEL_PLAYLISTS = QStringLiteral("/api/v1/channels/playlists");
 const QString API_HISTORY = QStringLiteral("/api/v1/auth/history");
 const QString API_COMMENTS = QStringLiteral("/api/v1/comments");
 const QString API_LIST_PLAYLISTS = QStringLiteral("/api/v1/auth/playlists");
@@ -27,6 +28,8 @@ const QString API_PREFERENCES = QStringLiteral("/api/v1/auth/preferences");
 const QString API_PLAYLISTS = QStringLiteral("/api/v1/playlists");
 const QString API_CHANNEL_INFO = QStringLiteral("/api/v1/channels");
 
+const QString COOKIE_KEY = QStringLiteral("cookie");
+
 using namespace QInvidious;
 using namespace Qt::StringLiterals;
 
@@ -35,11 +38,43 @@ InvidiousApi::InvidiousApi(QNetworkAccessManager *netManager, QObject *parent)
 {
 }
 
+bool InvidiousApi::isLoggedIn() const
+{
+    return m_cookie.has_value();
+}
+
+bool InvidiousApi::canLogIn() const
+{
+    // TODO: can we check if login is disabled, is that even an option?
+    return true;
+}
+
+void InvidiousApi::loadCredentials(const QString &prefix)
+{
+    if (auto cookie = getKeychainValue(prefix, COOKIE_KEY)) {
+        m_cookie = QNetworkCookie(cookie->toUtf8());
+    }
+    Q_EMIT credentialsChanged();
+}
+
+void InvidiousApi::saveCredentials(const QString &prefix)
+{
+    if (auto cookie = m_cookie) {
+        setKeychainValue(prefix, COOKIE_KEY, QString::fromUtf8(cookie->toRawForm()));
+    }
+}
+
+void InvidiousApi::wipeCredentials(const QString &prefix)
+{
+    wipeKeychainValue(prefix, COOKIE_KEY);
+    m_cookie = std::nullopt;
+    Q_EMIT credentialsChanged();
+}
+
 bool InvidiousApi::supportsFeature(AbstractApi::SupportedFeature feature)
 {
     switch (feature) {
     case PopularPage:
-        return true;
     case TrendingCategories:
         return true;
     }
@@ -47,11 +82,11 @@ bool InvidiousApi::supportsFeature(AbstractApi::SupportedFeature feature)
     return false;
 }
 
-QFuture<LogInResult> InvidiousApi::logIn(QStringView username, QStringView password)
+QFuture<LogInResult> InvidiousApi::logIn(const QString &username, const QString &password)
 {
     QUrlQuery params;
-    params.addQueryItem(QStringLiteral("email"), QString::fromUtf8(QUrl::toPercentEncoding(username.toString())));
-    params.addQueryItem(QStringLiteral("password"), QString::fromUtf8(QUrl::toPercentEncoding(password.toString())));
+    params.addQueryItem(QStringLiteral("email"), QString::fromUtf8(QUrl::toPercentEncoding(username)));
+    params.addQueryItem(QStringLiteral("password"), QString::fromUtf8(QUrl::toPercentEncoding(password)));
     params.addQueryItem(QStringLiteral("action"), QStringLiteral("signin"));
 
     QUrl url = apiUrl(API_LOGIN);
@@ -68,16 +103,15 @@ QFuture<LogInResult> InvidiousApi::logIn(QStringView username, QStringView passw
         const auto cookies = reply->header(QNetworkRequest::SetCookieHeader).value<QList<QNetworkCookie>>();
 
         if (!cookies.isEmpty()) {
-            m_credentials.setUsername(username);
-            m_credentials.setCookie(cookies.first());
+            m_cookie = cookies.first();
             Q_EMIT credentialsChanged();
-            return m_credentials;
+            return std::nullopt;
         }
         return std::pair(QNetworkReply::ContentAccessDenied, i18n("Username or password is wrong."));
     });
 }
 
-QFuture<VideoResult> InvidiousApi::requestVideo(QStringView videoId)
+QFuture<VideoResult> InvidiousApi::requestVideo(const QString &videoId)
 {
     return get<VideoResult>(QNetworkRequest(videoUrl(videoId)), [=](QNetworkReply *reply) -> VideoResult {
         if (auto doc = QJsonDocument::fromJson(reply->readAll()); !doc.isNull()) {
@@ -87,14 +121,23 @@ QFuture<VideoResult> InvidiousApi::requestVideo(QStringView videoId)
     });
 }
 
-QString InvidiousApi::resolveVideoUrl(QStringView videoId)
+QString InvidiousApi::resolveVideoUrl(const QString &videoId)
 {
     return QStringLiteral("ytdl://%1").arg(videoId);
 }
 
-QFuture<SearchListResult> InvidiousApi::requestSearchResults(const SearchParameters &parameters)
+QFuture<SearchListResult> InvidiousApi::requestSearchResults(const SearchParameters &parameters, Paginator *paginator)
 {
-    auto url = videoListUrl(Search, QStringLiteral(""), parameters.toQueryParameters());
+    QHash<QString, QString> searchParameters = parameters.toQueryParameters();
+
+    // Invidious uses pages for all of its video results
+    if (paginator != nullptr) {
+        paginator->setType(Paginator::Type::Page);
+        paginator->m_maxPage = std::numeric_limits<qsizetype>::max();
+        searchParameters.insert(QStringLiteral("page"), QString::number(paginator->m_page + 1));
+    }
+
+    auto url = videoListUrl(Search, QStringLiteral(""), searchParameters);
     auto request = QNetworkRequest(url);
 
     return get<SearchListResult>(std::move(request), [=](QNetworkReply *reply) -> SearchListResult {
@@ -114,12 +157,9 @@ QFuture<SearchListResult> InvidiousApi::requestSearchResults(const SearchParamet
     });
 }
 
-QFuture<VideoListResult> InvidiousApi::requestFeed(qint32 page)
+QFuture<VideoListResult> InvidiousApi::requestFeed(Paginator *paginator)
 {
-    QHash<QString, QString> parameters;
-    parameters.insert(QStringLiteral("page"), QString::number(page));
-
-    return requestVideoList(Feed, QStringLiteral(""), parameters);
+    return requestVideoList(Feed, QStringLiteral(""), {}, paginator);
 }
 
 QFuture<VideoListResult> InvidiousApi::requestTop()
@@ -127,7 +167,7 @@ QFuture<VideoListResult> InvidiousApi::requestTop()
     return requestVideoList(Top);
 }
 
-QFuture<VideoListResult> InvidiousApi::requestTrending(TrendingTopic topic)
+QFuture<VideoListResult> InvidiousApi::requestTrending(TrendingTopic topic, Paginator *paginator)
 {
     QHash<QString, QString> parameters;
     switch (topic) {
@@ -146,14 +186,12 @@ QFuture<VideoListResult> InvidiousApi::requestTrending(TrendingTopic topic)
     case Main:
         break;
     }
-    return requestVideoList(Trending, QStringLiteral(""), parameters);
+    return requestVideoList(Trending, QStringLiteral(""), parameters, paginator);
 }
 
-QFuture<VideoListResult> InvidiousApi::requestChannel(QStringView query, qint32 page)
+QFuture<VideoListResult> InvidiousApi::requestChannel(const QString &query, Paginator *paginator)
 {
-    QHash<QString, QString> parameters;
-    parameters.insert(QStringLiteral("page"), QString::number(page));
-    return requestVideoList(Channel, query.toString(), parameters);
+    return requestVideoList(Channel, query, {}, paginator);
 }
 
 QFuture<SubscriptionsResult> InvidiousApi::requestSubscriptions()
@@ -172,20 +210,23 @@ QFuture<SubscriptionsResult> InvidiousApi::requestSubscriptions()
     });
 }
 
-QFuture<Result> InvidiousApi::subscribeToChannel(QStringView channel)
+QFuture<Result> InvidiousApi::subscribeToChannel(const QString &channel)
 {
     return post<Result>(authenticatedNetworkRequest(subscribeUrl(channel)), {}, checkIsReplyOk);
 }
 
-QFuture<Result> InvidiousApi::unsubscribeFromChannel(QStringView channel)
+QFuture<Result> InvidiousApi::unsubscribeFromChannel(const QString &channel)
 {
     return deleteResource<Result>(authenticatedNetworkRequest(subscribeUrl(channel)), checkIsReplyOk);
 }
 
-QFuture<HistoryResult> InvidiousApi::requestHistory(qint32 page)
+QFuture<HistoryResult> InvidiousApi::requestHistory(Paginator *paginator)
 {
+    paginator->setType(Paginator::Type::Page);
+    paginator->m_maxPage = std::numeric_limits<qsizetype>::max();
+
     QUrlQuery parameters;
-    parameters.addQueryItem(QStringLiteral("page"), QString::number(page));
+    parameters.addQueryItem(QStringLiteral("page"), QString::number(paginator->m_page + 1));
 
     QUrl url = apiUrl(API_HISTORY);
     url.setQuery(parameters);
@@ -214,11 +255,13 @@ QFuture<Result> InvidiousApi::markUnwatched(const QString &videoId)
     return deleteResource<Result>(authenticatedNetworkRequest(apiUrl(API_HISTORY % u'/' % videoId)), checkIsReplyOk);
 }
 
-QFuture<CommentsResult> InvidiousApi::requestComments(const QString &videoId, const QString &continuation)
+QFuture<CommentsResult> InvidiousApi::requestComments(const QString &videoId, Paginator *paginator)
 {
+    paginator->setType(Paginator::Type::Continuation);
+
     QUrlQuery parameters;
-    if (!continuation.isEmpty()) {
-        parameters.addQueryItem(QStringLiteral("continuation"), continuation);
+    if (!paginator->m_continuation.isEmpty()) {
+        parameters.addQueryItem(QStringLiteral("continuation"), paginator->m_continuation);
     }
 
     QUrl url = apiUrl(API_COMMENTS % u'/' % videoId);
@@ -234,7 +277,10 @@ QFuture<CommentsResult> InvidiousApi::requestComments(const QString &videoId, co
                 Comment::fromJson(val.toObject(), comment);
                 return comment;
             });
-            return Comments{comments, doc[u"continuation"].toString()};
+
+            paginator->m_continuation = doc[u"continuation"].toString();
+
+            return comments;
         }
         return invalidJsonError();
     });
@@ -295,7 +341,7 @@ QFuture<VideoListResult> InvidiousApi::requestPlaylist(const QString &plid)
     });
 }
 
-QFuture<ChannelResult> InvidiousApi::requestChannelInfo(QStringView queryd)
+QFuture<ChannelResult> InvidiousApi::requestChannelInfo(const QString &queryd)
 {
     QUrl url = apiUrl(API_CHANNEL_INFO % u'/' % queryd);
 
@@ -309,6 +355,26 @@ QFuture<ChannelResult> InvidiousApi::requestChannelInfo(QStringView queryd)
     });
 }
 
+QFuture<PlaylistsResult> InvidiousApi::requestChannelPlaylists(const QString &channelId)
+{
+    QUrl url = apiUrl(API_CHANNEL_PLAYLISTS % u'/' % channelId);
+
+    return get<PlaylistsResult>(authenticatedNetworkRequest(std::move(url)), [=](QNetworkReply *reply) -> PlaylistsResult {
+        if (auto doc = QJsonDocument::fromJson(reply->readAll()); !doc.isNull()) {
+            const auto array = doc["playlists"_L1].toArray();
+
+            QList<Playlist> playlists;
+            std::transform(array.cbegin(), array.cend(), std::back_inserter(playlists), [](const QJsonValue &val) {
+                Playlist playlist;
+                Playlist::fromJson(val.toObject(), playlist);
+                return playlist;
+            });
+            return playlists;
+        }
+        return invalidJsonError();
+    });
+}
+
 QFuture<Result> InvidiousApi::addVideoToPlaylist(const QString &plid, const QString &videoId)
 {
     QUrl url = apiUrl(API_LIST_PLAYLISTS % u'/' % plid % u"/videos");
@@ -343,9 +409,27 @@ Result InvidiousApi::checkIsReplyOk(QNetworkReply *reply)
     return std::pair(QNetworkReply::InternalServerError, i18n("Server returned the status code %1", QString::number(status)));
 }
 
-QFuture<VideoListResult> InvidiousApi::requestVideoList(VideoListType queryType, const QString &urlExtension, const QHash<QString, QString> &parameters)
+QFuture<VideoListResult>
+InvidiousApi::requestVideoList(VideoListType queryType, const QString &urlExtension, const QHash<QString, QString> &parameters, Paginator *paginator)
 {
-    auto url = videoListUrl(queryType, urlExtension, parameters);
+    // TODO: we should really make the parameter copy, instead of doing it here?
+    QHash<QString, QString> finalParameters = parameters;
+
+    // Invidious uses pages for all of its video results
+    if (paginator != nullptr) {
+        paginator->setType(Paginator::Type::Page);
+
+        // Invidious doesn't support paginating through the Trending section.
+        if (queryType == Trending) {
+            paginator->m_maxPage = 0;
+        } else {
+            paginator->m_maxPage = std::numeric_limits<qsizetype>::max();
+            // TODO: Invidious pages start at 1. I wonder if we should bake this into the Paginator API?
+            finalParameters.insert(QStringLiteral("page"), QString::number(paginator->m_page + 1));
+        }
+    }
+
+    auto url = videoListUrl(queryType, urlExtension, finalParameters);
     // Feed requests require to be authenticated
     auto request = queryType == Feed ? authenticatedNetworkRequest(std::move(url)) : QNetworkRequest(url);
 
@@ -383,8 +467,8 @@ QFuture<VideoListResult> InvidiousApi::requestVideoList(VideoListType queryType,
 QNetworkRequest InvidiousApi::authenticatedNetworkRequest(QUrl &&url)
 {
     QNetworkRequest request(url);
-    if (!m_credentials.isAnonymous()) {
-        const QList<QNetworkCookie> cookies{m_credentials.cookie().value()};
+    if (isLoggedIn()) {
+        const QList<QNetworkCookie> cookies{m_cookie.value()};
         request.setHeader(QNetworkRequest::CookieHeader, QVariant::fromValue(cookies));
     }
     // some invidious instances redirect some calls using reverse proxies
@@ -404,7 +488,7 @@ QUrlQuery InvidiousApi::genericUrlQuery() const
     return query;
 }
 
-QUrl InvidiousApi::videoUrl(QStringView videoId) const
+QUrl InvidiousApi::videoUrl(const QString &videoId) const
 {
     return apiUrl(API_VIDEOS % u'/' % videoId);
 }
@@ -428,7 +512,7 @@ QUrl InvidiousApi::videoListUrl(VideoListType queryType, const QString &urlExten
         urlString.append(API_FEED);
         break;
     case Channel:
-        urlString.append(API_CHANNELS);
+        urlString.append(API_CHANNEL_VIDEOS);
         break;
     }
 
@@ -455,7 +539,7 @@ QUrl InvidiousApi::subscriptionsUrl() const
     return url;
 }
 
-QUrl InvidiousApi::subscribeUrl(QStringView channelId) const
+QUrl InvidiousApi::subscribeUrl(const QString &channelId) const
 {
     return apiUrl(API_SUBSCRIPTIONS % u'/' % channelId);
 }
