@@ -3,6 +3,7 @@
 
 #pragma once
 
+#include "config.h"
 #include "sourcemanager.h"
 #include "videocontroller.h"
 
@@ -40,6 +41,8 @@ public:
     Q_INVOKABLE void setApplicationProxy();
     Q_INVOKABLE void saveCurrentSource();
 
+    PlasmaTubeSettings *settings();
+
 Q_SIGNALS:
     void openVideo(const QString &id);
     void errorOccurred(const QString &errorText);
@@ -51,4 +54,6 @@ private:
 
     VideoController *m_controller = nullptr;
     SourceManager *m_sourceManager = nullptr;
+
+    PlasmaTubeSettings m_settings;
 };
